@@ -1,43 +1,29 @@
 """
 Open compressed files transparently.
 """
-from __future__ import print_function, division, absolute_import
+
+__all__ = ["xopen", "PipedGzipWriter", "PipedGzipReader", "__version__"]
 
 import gzip
 import sys
 import io
 import os
+import bz2
 import time
+import stat
 import signal
+import pathlib
 from subprocess import Popen, PIPE
 
 from ._version import version as __version__
 
 
-_PY3 = sys.version > '3'
-
-if not _PY3:
-    import bz2file as bz2
-else:
-    try:
-        import bz2
-    except ImportError:
-        bz2 = None
-
 try:
     import lzma
 except ImportError:
     lzma = None
 
 
-if _PY3:
-    basestring = str
-
-try:
-    import pathlib  # Exists in Python 3.4+
-except ImportError:
-    pathlib = None
-
 try:
     from os import fspath  # Exists in Python 3.6+
 except ImportError:
@@ -48,7 +34,7 @@ except ImportError:
         # path protocol
         if pathlib is not None and isinstance(path, pathlib.Path):
             return str(path)
-        if not isinstance(path, basestring):
+        if not isinstance(path, str):
             raise TypeError("path must be a string")
         return path
 
@@ -71,7 +57,7 @@ def _available_cpu_count():
             res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
             if res > 0:
                 return res
-    except IOError:
+    except OSError:
         pass
     try:
         import multiprocessing
@@ -80,11 +66,12 @@ def _available_cpu_count():
         return 1
 
 
-class Closing(object):
+class Closing:
     """
     Inherit from this class and implement a close() method to offer context
     manager functionality.
     """
+
     def __enter__(self):
         return self
 
@@ -94,7 +81,7 @@ class Closing(object):
     def __del__(self):
         try:
             self.close()
-        except:
+        except Exception:
             pass
 
 
@@ -119,7 +106,8 @@ class PipedGzipWriter(Closing):
             at four to avoid creating too many threads. Use 0 to let pigz use all available cores.
         """
         if mode not in ('w', 'wt', 'wb', 'a', 'at', 'ab'):
-            raise ValueError("Mode is '{0}', but it must be 'w', 'wt', 'wb', 'a', 'at' or 'ab'".format(mode))
+            raise ValueError(
+                "Mode is '{}', but it must be 'w', 'wt', 'wb', 'a', 'at' or 'ab'".format(mode))
 
         # TODO use a context manager
         self.outfile = open(path, mode)
@@ -127,7 +115,32 @@ class PipedGzipWriter(Closing):
         self.closed = False
         self.name = path
 
-        kwargs = dict(stdin=PIPE, stdout=self.outfile, stderr=self.devnull)
+        if threads is None:
+            threads = min(_available_cpu_count(), 4)
+        try:
+            self.process, self.program = self._open_process(
+                mode, compresslevel, threads, self.outfile, self.devnull)
+        except OSError:
+            self.outfile.close()
+            self.devnull.close()
+            raise
+
+        if 'b' not in mode:
+            self._file = io.TextIOWrapper(self.process.stdin)
+        else:
+            self._file = self.process.stdin
+
+    @staticmethod
+    def _open_process(mode, compresslevel, threads, outfile, devnull):
+        pigz_args = ['pigz']
+        if threads != 0:
+            pigz_args += ['-p', str(threads)]
+        extra_args = []
+        if 'w' in mode and compresslevel != 6:
+            extra_args += ['-' + str(compresslevel)]
+
+        kwargs = dict(stdin=PIPE, stdout=outfile, stderr=devnull)
+
         # Setting close_fds to True in the Popen arguments is necessary due to
         # <http://bugs.python.org/issue12786>.
         # However, close_fds is not supported on Windows. See
@@ -135,36 +148,14 @@ class PipedGzipWriter(Closing):
         if sys.platform != 'win32':
             kwargs['close_fds'] = True
 
-        if 'w' in mode and compresslevel != 6:
-            extra_args = ['-' + str(compresslevel)]
-        else:
-            extra_args = []
-
-        pigz_args = ['pigz']
-        if threads is None:
-            threads = min(_available_cpu_count(), 4)
-        if threads != 0:
-            pigz_args += ['-p', str(threads)]
         try:
-            self.process = Popen(pigz_args + extra_args, **kwargs)
-            self.program = 'pigz'
-        except OSError:
+            process = Popen(pigz_args + extra_args, **kwargs)
+            program = 'pigz'
+        except OSError:  # TODO Use FileNotFound instead (Python 3)
             # pigz not found, try regular gzip
-            try:
-                self.process = Popen(['gzip'] + extra_args, **kwargs)
-                self.program = 'gzip'
-            except (IOError, OSError):
-                self.outfile.close()
-                self.devnull.close()
-                raise
-        except IOError:  # TODO IOError is the same as OSError on Python 3.3
-            self.outfile.close()
-            self.devnull.close()
-            raise
-        if _PY3 and 'b' not in mode:
-            self._file = io.TextIOWrapper(self.process.stdin)
-        else:
-            self._file = self.process.stdin
+            process = Popen(['gzip'] + extra_args, **kwargs)
+            program = 'gzip'
+        return process, program
 
     def write(self, arg):
         self._file.write(arg)
@@ -178,7 +169,8 @@ class PipedGzipWriter(Closing):
         self.outfile.close()
         self.devnull.close()
         if retcode != 0:
-            raise IOError("Output {0} process terminated with exit code {1}".format(self.program, retcode))
+            raise OSError(
+                "Output {} process terminated with exit code {}".format(self.program, retcode))
 
     def __iter__(self):
         return self
@@ -200,7 +192,7 @@ class PipedGzipReader(Closing):
         Raise an OSError when pigz could not be found.
         """
         if mode not in ('r', 'rt', 'rb'):
-            raise ValueError("Mode is '{0}', but it must be 'r', 'rt' or 'rb'".format(mode))
+            raise ValueError("Mode is '{}', but it must be 'r', 'rt' or 'rb'".format(mode))
 
         pigz_args = ['pigz', '-cd', path]
 
@@ -217,14 +209,11 @@ class PipedGzipReader(Closing):
 
         self.process = Popen(pigz_args, stdout=PIPE, stderr=PIPE)
         self.name = path
-        if _PY3 and 'b' not in mode:
+        if 'b' not in mode:
             self._file = io.TextIOWrapper(self.process.stdout)
         else:
             self._file = self.process.stdout
-        if _PY3:
-            self._stderr = io.TextIOWrapper(self.process.stderr)
-        else:
-            self._stderr = self.process.stderr
+        self._stderr = io.TextIOWrapper(self.process.stderr)
         self.closed = False
         # Give the subprocess a little bit of time to report any errors (such as
         # a non-existing file)
@@ -243,10 +232,15 @@ class PipedGzipReader(Closing):
         else:
             allow_sigterm = False
         self.process.wait()
+        self._file.close()
         self._raise_if_error(allow_sigterm=allow_sigterm)
+        self._stderr.close()
 
     def __iter__(self):
-        return self._file
+        return self
+
+    def __next__(self):
+        return self._file.__next__()
 
     def _raise_if_error(self, allow_sigterm=False):
         """
@@ -260,7 +254,9 @@ class PipedGzipReader(Closing):
             and not (allow_sigterm and retcode == -signal.SIGTERM)
         ):
             message = self._stderr.read().strip()
-            raise IOError("{} (exit code {})".format(message, retcode))
+            self._file.close()
+            self._stderr.close()
+            raise OSError("{} (exit code {})".format(message, retcode))
 
     def read(self, *args):
         return self._file.read(*args)
@@ -278,12 +274,7 @@ class PipedGzipReader(Closing):
         return self._file.peek(n)
 
     def readable(self):
-        if _PY3:
-            return self._file.readable()
-        else:
-            return NotImplementedError(
-                "Python 2 does not support the readable() method."
-            )
+        return self._file.readable()
 
     def writable(self):
         return self._file.writable()
@@ -296,23 +287,11 @@ def _open_stdin_or_out(mode):
     # Do not return sys.stdin or sys.stdout directly as we want the returned object
     # to be closable without closing sys.stdout.
     std = dict(r=sys.stdin, w=sys.stdout)[mode[0]]
-    if not _PY3:
-        # Enforce str type on Python 2
-        # Note that io.open is slower than regular open() on Python 2.7, but
-        # it appears to be the only API that has a closefd parameter.
-        mode = mode[0] + 'b'
-    return io.open(std.fileno(), mode=mode, closefd=False)
+    return open(std.fileno(), mode=mode, closefd=False)
 
 
 def _open_bz2(filename, mode):
-    if bz2 is None:
-        raise ImportError("Cannot open bz2 files: The bz2 module is not available")
-    if _PY3:
-        return bz2.open(filename, mode)
-    else:
-        if mode[0] == 'a':
-            raise ValueError("mode '{0}' not supported with BZ2 compression".format(mode))
-        return bz2.BZ2File(filename, mode)
+    return bz2.open(filename, mode)
 
 
 def _open_xz(filename, mode):
@@ -323,37 +302,56 @@ def _open_xz(filename, mode):
 
 
 def _open_gz(filename, mode, compresslevel, threads):
-    if sys.version_info[:2] == (2, 7):
-        buffered_reader = io.BufferedReader
-        buffered_writer = io.BufferedWriter
-    else:
-        buffered_reader = lambda x: x
-        buffered_writer = lambda x: x
-    if _PY3:
-        exc = FileNotFoundError  # was introduced in Python 3.3
-    else:
-        exc = OSError
+    if threads != 0:
+        try:
+            if 'r' in mode:
+                return PipedGzipReader(filename, mode, threads=threads)
+            else:
+                return PipedGzipWriter(filename, mode, compresslevel, threads=threads)
+        except FileNotFoundError:
+            pass  # We try without threads.
 
     if 'r' in mode:
-        def open_with_threads():
-            return PipedGzipReader(filename, mode, threads=threads)
-
-        def open_without_threads():
-            return buffered_reader(gzip.open(filename, mode))
+        return gzip.open(filename, mode)
     else:
-        def open_with_threads():
-            return PipedGzipWriter(filename, mode, compresslevel, threads=threads)
+        return gzip.open(filename, mode, compresslevel=compresslevel)
 
-        def open_without_threads():
-            return buffered_writer(gzip.open(filename, mode, compresslevel=compresslevel))
 
-    if threads == 0:
-        return open_without_threads()
+def _detect_format_from_content(filename):
+    """
+    Attempts to detect file format from the content by reading the first
+    6 bytes. Returns None if no format could be detected.
+    """
     try:
-        return open_with_threads()
-    except exc:
-        # pigz is not installed, use fallback
-        return open_without_threads()
+        if stat.S_ISREG(os.stat(filename).st_mode):
+            with open(filename, "rb") as fh:
+                bs = fh.read(6)
+            if bs[:2] == b'\x1f\x8b':
+                # https://tools.ietf.org/html/rfc1952#page-6
+                return "gz"
+            elif bs[:3] == b'\x42\x5a\x68':
+                # https://en.wikipedia.org/wiki/List_of_file_signatures
+                return "bz2"
+            elif bs[:6] == b'\xfd\x37\x7a\x58\x5a\x00':
+                # https://tukaani.org/xz/xz-file-format.txt
+                return "xz"
+    except OSError:
+        return None
+
+
+def _detect_format_from_extension(filename):
+    """
+    Attempts to detect file format from the filename extension.
+    Returns None if no format could be detected.
+    """
+    if filename.endswith('.bz2'):
+        return "bz2"
+    elif filename.endswith('.xz'):
+        return "xz"
+    elif filename.endswith('.gz'):
+        return "gz"
+    else:
+        return None
 
 
 def xopen(filename, mode='r', compresslevel=6, threads=None):
@@ -387,24 +385,23 @@ def xopen(filename, mode='r', compresslevel=6, threads=None):
     if mode in ('r', 'w', 'a'):
         mode += 't'
     if mode not in ('rt', 'rb', 'wt', 'wb', 'at', 'ab'):
-        raise ValueError("mode '{0}' not supported".format(mode))
-    if not _PY3:
-        mode = mode[0]
+        raise ValueError("Mode '{}' not supported".format(mode))
     filename = fspath(filename)
     if compresslevel not in range(1, 10):
         raise ValueError("compresslevel must be between 1 and 9")
 
     if filename == '-':
         return _open_stdin_or_out(mode)
-    elif filename.endswith('.bz2'):
-        return _open_bz2(filename, mode)
-    elif filename.endswith('.xz'):
-        return _open_xz(filename, mode)
-    elif filename.endswith('.gz'):
+
+    detected_format = _detect_format_from_extension(filename)
+    if detected_format is None and "w" not in mode:
+        detected_format = _detect_format_from_content(filename)
+
+    if detected_format == "gz":
         return _open_gz(filename, mode, compresslevel, threads)
+    elif detected_format == "xz":
+        return _open_xz(filename, mode)
+    elif detected_format == "bz2":
+        return _open_bz2(filename, mode)
     else:
-        # Python 2.6 and 2.7 have io.open, which we could use to make the returned
-        # object consistent with the one returned in Python 3, but reading a file
-        # with io.open() is 100 times slower (!) on Python 2.6, and still about
-        # three times slower on Python 2.7 (tested with "for _ in io.open(path): pass")
         return open(filename, mode)
