@@ -1,14 +1,11 @@
-# coding: utf-8
-from __future__ import print_function, division, absolute_import
-
 import io
 import os
 import random
-import sys
 import signal
 import time
-from contextlib import contextmanager
 import pytest
+from pathlib import Path
+
 from xopen import xopen, PipedGzipReader, PipedGzipWriter
 
 
@@ -25,11 +22,6 @@ files = [base + ext for ext in extensions]
 CONTENT_LINES = ['Testing, testing ...\n', 'The second line.\n']
 CONTENT = ''.join(CONTENT_LINES)
 
-# File extensions for which appending is supported
-append_extensions = extensions[:]
-if sys.version_info[0] == 2:
-    append_extensions.remove(".bz2")
-
 
 @pytest.fixture(params=extensions)
 def ext(request):
@@ -112,11 +104,20 @@ def test_pipedgzipreader_readinto():
         assert b[:length] == content
 
 
-if sys.version_info[0] != 2:
-    def test_pipedgzipreader_textiowrapper():
-        with PipedGzipReader("tests/file.txt.gz", "rb") as f:
-            wrapped = io.TextIOWrapper(f)
-            assert wrapped.read() == CONTENT
+def test_pipedgzipreader_textiowrapper():
+    with PipedGzipReader("tests/file.txt.gz", "rb") as f:
+        wrapped = io.TextIOWrapper(f)
+        assert wrapped.read() == CONTENT
+
+
+def test_detect_gzip_file_format_from_content():
+    with xopen("tests/file.txt.gz.test", "rb") as fh:
+        assert fh.readline() == CONTENT_LINES[0].encode("utf-8")
+
+
+def test_detect_bz2_file_format_from_content():
+    with xopen("tests/file.txt.bz2.test", "rb") as fh:
+        assert fh.readline() == CONTENT_LINES[0].encode("utf-8")
 
 
 def test_readline(fname):
@@ -141,6 +142,20 @@ def test_readline_text_pipedgzipreader():
         assert f.readline() == CONTENT_LINES[0]
 
 
+@pytest.mark.parametrize("threads", [None, 1, 2])
+def test_pipedgzipreader_iter(threads):
+    with PipedGzipReader("tests/file.txt.gz", mode="r", threads=threads) as f:
+        lines = list(f)
+        assert lines[0] == CONTENT_LINES[0]
+
+
+def test_next(fname):
+    with xopen(fname, "rt") as f:
+        _ = next(f)
+        line2 = next(f)
+        assert line2 == 'The second line.\n', fname
+
+
 def test_xopen_has_iter_method(ext, tmpdir):
     path = str(tmpdir.join("out" + ext))
     with xopen(path, mode='w') as f:
@@ -152,6 +167,16 @@ def test_pipedgzipwriter_has_iter_method(tmpdir):
         assert hasattr(f, '__iter__')
 
 
+def test_iter_without_with(fname):
+    it = iter(xopen(fname, "rt"))
+    assert CONTENT_LINES[0] == next(it)
+
+
+def test_pipedgzipreader_iter_without_with():
+    it = iter(PipedGzipReader("tests/file.txt.gz"))
+    assert CONTENT_LINES[0] == next(it)
+
+
 @pytest.mark.parametrize("mode", ["rb", "rt"])
 def test_pipedgzipreader_close(large_gzip, mode):
     with PipedGzipReader(large_gzip, mode=mode) as f:
@@ -160,7 +185,6 @@ def test_pipedgzipreader_close(large_gzip, mode):
     # The subprocess should be properly terminated now
 
 
-@pytest.mark.skipif(sys.version_info < (3, ), reason="Python 3 needed")
 def test_partial_gzip_iteration_closes_correctly(large_gzip):
     class LineReader:
         def __init__(self, file):
@@ -168,8 +192,7 @@ def test_partial_gzip_iteration_closes_correctly(large_gzip):
 
         def __iter__(self):
             wrapper = io.TextIOWrapper(self.file)
-            for line in wrapper:
-                yield line
+            yield from wrapper
 
     f = LineReader(large_gzip)
     next(iter(f))
@@ -178,25 +201,25 @@ def test_partial_gzip_iteration_closes_correctly(large_gzip):
 
 def test_nonexisting_file(ext):
     with pytest.raises(IOError):
-        with xopen('this-file-does-not-exist' + ext) as f:
+        with xopen('this-file-does-not-exist' + ext):
             pass  # pragma: no cover
 
 
 def test_write_to_nonexisting_dir(ext):
     with pytest.raises(IOError):
-        with xopen('this/path/does/not/exist/file.txt' + ext, 'w') as f:
+        with xopen('this/path/does/not/exist/file.txt' + ext, 'w'):
             pass  # pragma: no cover
 
 
 def test_invalid_mode():
     with pytest.raises(ValueError):
-        with xopen("tests/file.txt.gz", mode="hallo") as f:
+        with xopen("tests/file.txt.gz", mode="hallo"):
             pass  # pragma: no cover
 
 
 def test_filename_not_a_string():
     with pytest.raises(TypeError):
-        with xopen(123, mode="r") as f:
+        with xopen(123, mode="r"):
             pass  # pragma: no cover
 
 
@@ -208,11 +231,11 @@ def test_invalid_compression_level(tmpdir):
     assert "between 1 and 9" in e.value.args[0]
 
 
-@pytest.mark.parametrize("aext", append_extensions)
-def test_append(aext, tmpdir):
-    text = "AB".encode("utf-8")
+@pytest.mark.parametrize("ext", extensions)
+def test_append(ext, tmpdir):
+    text = b"AB"
     reference = text + text
-    path = str(tmpdir.join("the-file" + aext))
+    path = str(tmpdir.join("the-file" + ext))
     with xopen(path, "ab") as f:
         f.write(text)
     with xopen(path, "ab") as f:
@@ -224,11 +247,11 @@ def test_append(aext, tmpdir):
         assert appended == reference
 
 
-@pytest.mark.parametrize("aext", append_extensions)
-def test_append_text(aext, tmpdir):
+@pytest.mark.parametrize("ext", extensions)
+def test_append_text(ext, tmpdir):
     text = "AB"
     reference = text + text
-    path = str(tmpdir.join("the-file" + aext))
+    path = str(tmpdir.join("the-file" + ext))
     with xopen(path, "at") as f:
         f.write(text)
     with xopen(path, "at") as f:
@@ -309,18 +332,17 @@ def test_write_pigz_threads(tmpdir):
         assert f.read() == 'hello'
 
 
-if sys.version_info[0] >= 3:
-    def test_read_gzip_no_threads():
-        import gzip
-        with xopen("tests/hello.gz", "rb", threads=0) as f:
-            assert isinstance(f, gzip.GzipFile), f
+def test_read_gzip_no_threads():
+    import gzip
+    with xopen("tests/hello.gz", "rb", threads=0) as f:
+        assert isinstance(f, gzip.GzipFile), f
 
 
-    def test_write_gzip_no_threads(tmpdir):
-        import gzip
-        path = str(tmpdir.join("out.gz"))
-        with xopen(path, "wb", threads=0) as f:
-            assert isinstance(f, gzip.GzipFile), f
+def test_write_gzip_no_threads(tmpdir):
+    import gzip
+    path = str(tmpdir.join("out.gz"))
+    with xopen(path, "wb", threads=0) as f:
+        assert isinstance(f, gzip.GzipFile), f
 
 
 def test_write_stdout():
@@ -339,30 +361,36 @@ def test_write_stdout_contextmanager():
     print("Still there?")
 
 
-if sys.version_info[:2] >= (3, 4):
-    # pathlib was added in Python 3.4
-    from pathlib import Path
+def test_read_pathlib(fname):
+    path = Path(fname)
+    with xopen(path, mode='rt') as f:
+        assert f.read() == CONTENT
 
-    def test_read_pathlib(fname):
-        path = Path(fname)
-        with xopen(path, mode='rt') as f:
-            assert f.read() == CONTENT
 
-    def test_read_pathlib_binary(fname):
-        path = Path(fname)
-        with xopen(path, mode='rb') as f:
-            assert f.read() == bytes(CONTENT, 'ascii')
+def test_read_pathlib_binary(fname):
+    path = Path(fname)
+    with xopen(path, mode='rb') as f:
+        assert f.read() == bytes(CONTENT, 'ascii')
 
-    def test_write_pathlib(ext, tmpdir):
-        path = Path(str(tmpdir)) / ('hello.txt' + ext)
-        with xopen(path, mode='wt') as f:
-            f.write('hello')
-        with xopen(path, mode='rt') as f:
-            assert f.read() == 'hello'
 
-    def test_write_pathlib_binary(ext, tmpdir):
-        path = Path(str(tmpdir)) / ('hello.txt' + ext)
-        with xopen(path, mode='wb') as f:
-            f.write(b'hello')
-        with xopen(path, mode='rb') as f:
-            assert f.read() == b'hello'
+def test_write_pathlib(ext, tmpdir):
+    path = Path(str(tmpdir)) / ('hello.txt' + ext)
+    with xopen(path, mode='wt') as f:
+        f.write('hello')
+    with xopen(path, mode='rt') as f:
+        assert f.read() == 'hello'
+
+
+def test_write_pathlib_binary(ext, tmpdir):
+    path = Path(str(tmpdir)) / ('hello.txt' + ext)
+    with xopen(path, mode='wb') as f:
+        f.write(b'hello')
+    with xopen(path, mode='rb') as f:
+        assert f.read() == b'hello'
+
+
+# lzma doesn’t work on PyPy3 at the moment
+if lzma is not None:
+    def test_detect_xz_file_format_from_content():
+        with xopen("tests/file.txt.xz.test", "rb") as fh:
+            assert fh.readline() == CONTENT_LINES[0].encode("utf-8")
